/* Copyright (C) 2017 Advanced Digital Sciences Centre
 * 
 * This is a software-based attestation demo code developed at Advanced Digital Sciences Center.
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * It is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @author: Toby Zhou.Bin
 */


/* 
 * File:   Prover.c
 * Author: Zhou Bin
 *
 * Created on 14 Sep, 2017, 4:20 PM
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>
#include "Utils.h"
#include "MemoryStride.h"

#define BUF_SIZE 1024
#define MESSAGE_SIZE 66
#define MESSAGE_HEADER_SIZE 6
#define MESSAGE_CONTENT_SIZE 60  // seed + iter
#define MESSAGE_SEED_SIZE 52  // 13 seeds, 1 seed = 4 bytes
#define MESSAGE_ITER_SIZE 8
#define SEED_NUMBER 13
#define SEED_SIZE 4
#define CHECKSUM_SIZE 48


void warnAndExit(char *msg) {
    colorLog(KRED, msg);
    exit(0);
}


void sendFeedback (int sockNo, char* buf) {
    int writeRes = write(sockNo, buf, strlen(buf));
    if (writeRes < 0) {
        warnAndExit("Unable to send packets to client. Aborting ...\n");
    } else {
        printf("I've replied: %s\n\n", buf);
    }
    return;
}


int checkAttestPacket(unsigned char * buf) {
    // Only check the first 6 bytes of message received.
    if (buf[0] == 0x68 && buf[1] == 0x04 && buf[2] == 0x03
        && buf[3] == 0x66 && buf[4] == 0x69 && buf[5] == 0x99) {
            return 1;
    }
    return 0;
}


void process(int bySockNo, unsigned char * msg, int msgSize, char* memoryImg) {
    int i=0, j=0; // for loop

    if(msgSize != MESSAGE_CONTENT_SIZE)  {
        char* buf = "Invalid message received...\n";
        sendFeedback(bySockNo, buf);
        return ;
    }

    // Read seeds & iteration
    unsigned long seeds[SEED_NUMBER];
    unsigned char seed[SEED_SIZE + 1]; // +1 to store '\0'
    unsigned char iter[MESSAGE_ITER_SIZE+1];  // +1 to store '\0'
    unsigned long iteration;

    // Get Seeds
    unsigned char tempChar;
    for(i=0; i<SEED_NUMBER; i++) {
        for (j=0; j<SEED_SIZE; j++) {
            tempChar = msg[i * SEED_SIZE + j];
            if(isValidValue(tempChar)) 
                seed[j] = tempChar;
            else {
                char* buf = "Message contains invalid characters. Please check...\n";
                sendFeedback(bySockNo, buf);
                return ;
            }
        }
        seed[SEED_SIZE] = '\0';
        seeds[i] = stringToHex(seed, SEED_SIZE);
    }

    // Get iteration
    for(i=0; i<MESSAGE_ITER_SIZE; i++) {
        tempChar = msg[MESSAGE_SEED_SIZE + i];
        if(isValidValue(tempChar)) {
            iter[i] = tempChar;
        } else {
            char* buf = "Message contains invalid characters. Please check...\n";
            sendFeedback(bySockNo, buf);
            return ;
        }
    }
    iter[MESSAGE_ITER_SIZE] = '\0';
    iteration = stringToHex(iter, MESSAGE_ITER_SIZE);


    // Check values
    for(i=0; i<SEED_NUMBER; i++)
        printf("Seed %d: %02lx \n", i, seeds[i]);
    printf("Iteration: %lu\n", iteration); 

    
    // Call MemoryStride
    // Inputs: memory log, 13 seeds, iteration
    char simulatedMainRes[CHECKSUM_SIZE] = {0};
    strideMemory(simulatedMainRes, memoryImg, seeds, iteration);
    int writeRes = write(bySockNo, simulatedMainRes, CHECKSUM_SIZE);
    if(writeRes < 0) 
        warnAndExit("Unable to send packets to client. Aborting ...\n");
    else
        printf("Sent back computation result. \n");

    return;
}



int main(int argc, char *argv[])
{

    // Get Configurations: port and memoryImg
    char memoryImg[BUF_SIZE];
    int portNum;
    getProverConfig(memoryImg, &portNum);
    printf("Configs: %s\n%d\n\n", memoryImg, portNum);

    // Create Socket.
    int sockfd = createSocket();

    struct sockaddr_in serv_addr;
    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr.s_addr = INADDR_ANY;
    serv_addr.sin_port = htons(portNum);

    if (bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
        warnAndExit("Unable to bind socket. Aborting ...\n");

    listen(sockfd,5);

    // Loop for continuous requests
    char buffer[BUF_SIZE];
    int newsockfd=-1, clilen=-1, readRes=-1;
    struct sockaddr_in  cli_addr;
    clilen = sizeof(cli_addr);
    while(1) {
        // Accept connections with client
        newsockfd = accept(sockfd, (struct sockaddr *) &cli_addr, &clilen);
        if (newsockfd < 0)
            warnAndExit("Unable to accept client. Aborting ...\n");


        // Read packets from client
        bzero(buffer,BUF_SIZE);
        readRes = read(newsockfd, buffer, BUF_SIZE-1);
        if (readRes < 0)
            warnAndExit("Unable to read packets from client. Aborting ...\n");

        // Print packet message
        printf("\nReceived the message!\n - size: %d\n - Message: %s\n - Message(hex):", readRes, buffer);
        displayInHex((unsigned char *)buffer, readRes);

        // MemoryStride
        if (checkAttestPacket((unsigned char *)buffer)) {
            process(newsockfd, (unsigned char *)buffer + MESSAGE_HEADER_SIZE,
                    readRes - MESSAGE_HEADER_SIZE, memoryImg);
        } else {
            char* buf = "This is not Memory Stride Package.";
            sendFeedback(newsockfd, buf);
        }
    }

    return 0; 
}