/* Copyright (C) 2017 Advanced Digital Sciences Centre
 * 
 * This is a software-based attestation demo code developed at Advanced Digital Sciences Center.
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * It is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @author: Toby Zhou.Bin
 */

/* 
 * File:   Utils.c
 * Author: Zhou Bin
 *
 * Created on 14 Sep, 2017, 5:10 PM
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>

#include "Utils.h"


#define BASE_16 16
#define BUFSIZE 1024
#define VERIFIER_CONFIG "Verifier.conf"
#define PROVER_CONFIG "Prover.conf"
#define CONFIG_MEMORY "Memory_Image"
#define CONFIG_IP "Prover_IP"
#define CONFIG_PORT "Prover_Port"
#define CONFIG_ITERATION "Iterations"
#define CONFIG_TIME_THRESHOLD "Time_Threshold"


void colorLog(char * color, unsigned char * content){
    printf("%s", color);
    printf("%s", content);
    printf("%s",KNRM);
}


void displayInHex(const unsigned char *str, int n) {
    int i=0; // for loop

    for(i=0; i<n; i++)
        printf("%02x", (unsigned int) *str++);
    
    printf("\n");
}


unsigned int charToNum (unsigned char c) {
    if ('0' <= c && c <= '9') return c - '0';
    if ('a' <= c && c <= 'f') return 10 + c - 'a';
    if ('A' <= c && c <= 'F') return 10 + c - 'A';
    return 0;
}

unsigned long stringToHex(unsigned char * value, int size) {
    int i=0; // for loop
    unsigned long result = 0;

    for (i = size-1; i >= 0; i--) {
        result += pow(16, i) * charToNum(*(value + ((size-1) -i)));
    }
    return result;
}

/* Hex to LL Int converter*/
unsigned long long int hex2int(char *a, unsigned int len) {
    int i;
    unsigned long int val = 0;

    for (i = 0; i < len; i++)
        if (a[i] <= 57)
            val += (a[i] - 48)*(1 << (4 * (len - 1 - i)));
        else
            val += (a[i] - 55)*(1 << (4 * (len - 1 - i)));

    return val;
}

/* Implement ARM ROR Rotate Right instruction*/
unsigned long _rotr(unsigned long value, int shift) {
    if ((shift &= 31) == 0)
        return value;
    return (value >> shift) | (value << (32 - shift));
}


float timedifference_msec(struct timeval t0, struct timeval t1) {
    return (t1.tv_sec - t0.tv_sec) * 1000.0f + (t1.tv_usec - t0.tv_usec) / 1000.0f;
}


int isValidValue(char value) {
    if((value >= '0' && value <= '9') ||
        (value >= 'A' && value <= 'F')) 
        return 1;
    return 0;
}


void getVerifierConfig(char* memoryImg, char* IP, int* portNum, int* timeThreshold, unsigned long* iterations) {
    FILE *fp = fopen(VERIFIER_CONFIG, "rt");
    if(!fp) {
        printf("The file %s is missing. Please check and put it in the same directory as verifier.\n",
               VERIFIER_CONFIG);
        exit(0);
    }
    
    char line[BUFSIZE];
    bool hasMem=false, hasIP=false, hasPort=false, hasThreshold=false, hasIteration=false;
    while(fgets(line, sizeof(line), fp)) {
        
        // Get line, then split into name and value. Remove new line char if any
        char* firstColon  = strchr(line, ':');
        *firstColon = '\0';
        char* name = line;
        char* value = firstColon + 1;

        char* newLineChar = strchr(value, '\n');
        if(newLineChar!=NULL) {
            *newLineChar = '\0';
        }

        char* newLineCharLinux = strchr(value, '\r'); // special case for linux: '\r\n'
        if(newLineCharLinux!=NULL) {
            *newLineCharLinux = '\0';
        }
        
        if(strcmp(name, CONFIG_MEMORY) == 0) {
            strncpy(memoryImg, value, strlen(value)+1);  // +1 to copy the '\0' char

            FILE *memoryFp = fopen(memoryImg, "rt");
            if(!memoryFp) {
                printf("The memory log doesn't exist. Please check and update config file.\n%s\n", memoryImg);
                exit(0);
            }

            hasMem = true;
            fclose(memoryFp);
        } else if(strcmp(name, CONFIG_IP) == 0) {
            strncpy(IP, value, strlen(value)+1);  // +1 to copy the '\0' char
            hasIP = true;
        } else if(strcmp(name, CONFIG_PORT) == 0) {
            *portNum = atoi(value);
            hasPort = true;
        } else if(strcmp(name, CONFIG_ITERATION) == 0) {
            char* ptr;
            *iterations = strtoul(value, &ptr, BASE_16);
            hasIteration = true;
        } else if(strcmp(name, CONFIG_TIME_THRESHOLD) == 0) {
            *timeThreshold = atoi(value);
            hasThreshold = true;
        } else {
            printf("Unexpected configurations. Please check.\n");
            exit(0);
        }
    }

    if(!(hasMem && hasIP && hasIteration && hasPort && hasThreshold)) {
        printf("Missing configurations. Must include %s, %s, %s, %s, %s \n", 
               CONFIG_MEMORY, CONFIG_IP, CONFIG_PORT, CONFIG_ITERATION, CONFIG_TIME_THRESHOLD);
        exit(0);
    }

    fclose(fp);
}


void getProverConfig(char* memoryImg, int* portNum) {
    FILE *fp = fopen(PROVER_CONFIG, "rt");
    if(!fp) {
        printf("The file %s is missing. Please check and put it in the same directory as Prover.\n",
               PROVER_CONFIG);
        exit(0);
    }

    char line[BUFSIZE];
    bool hasMem=false, hasPort=false;
    while(fgets(line, sizeof(line), fp)) {
        
        // Get line, then split into name and value. Remove new line char if any
        char* firstColon  = strchr(line, ':');
        *firstColon = '\0';
        char* name = line;
        char* value = firstColon + 1;

        char* newLineChar = strchr(value, '\n');
        if(newLineChar!=NULL) {
            *newLineChar = '\0';
        }

        char* newLineCharLinux = strchr(value, '\r'); // special case for linux: '\r\n'
        if(newLineCharLinux!=NULL) {
            *newLineCharLinux = '\0';
        } 
        
        if(strcmp(name, CONFIG_MEMORY) == 0) {
            strncpy(memoryImg, value, strlen(value)+1);  // +1 to copy the '\0' char

            FILE *memoryFp = fopen(memoryImg, "rt");
            if(!memoryFp) {
                printf("The memory log doesn't exist. Please check and update config file.\n%s\n", memoryImg);
                exit(0);
            }

            hasMem = true;
            fclose(memoryFp);
        } else if(strcmp(name, CONFIG_PORT) == 0) {
            *portNum = atoi(value);
            hasPort = true;
        } else {
            printf("Unexpected configurations. Please check.\n");
            exit(0);
        }
    }

    if(!(hasMem && hasPort)) {
        printf("Missing configurations. Must include %s, %s \n", 
               CONFIG_MEMORY, CONFIG_PORT);
        exit(0);
    }

    fclose(fp);
}


int createSocket() {
    /* Create a socket connection for communications */
    /* Try a maximum of 4 times before exiting */
    int sockfd = -1, tries = 0;
    while (tries < 4 && sockfd < 0) {
        sockfd = socket(AF_INET, SOCK_STREAM, 0);
        tries++;
        if (sockfd < 0) {
            printf("ERROR in opening  a socket for communication . Retrying attempt %d \n", tries);
        } else {
            printf("Successfully create socket.\n");
            break;
        }
        if (tries == 4) {
            colorLog(KRED, "Unable to open socket after 4 attempts. Aborting ...\n");
            exit(0);
        }
    } 
    return sockfd;
}


struct hostent * findProver(char * proverIP) {
    struct hostent *Prover;
    Prover = gethostbyname(proverIP);
    if (Prover == NULL) {
        colorLog(KRED, "ERROR Unable to resolve the IP address of the Prover. Aborting ...\n");
        exit(0);
    } else {
        printf("Successfully find the Prover's IP address: ");
        printf("%s\n", Prover->h_name);
    }
    return Prover;
}


int connectProver(int sockfd, struct sockaddr * serveraddr, int serveraddr_size) {
    int confd = -1, tries = 0;
    
    while (tries < 4 && confd < 0) {
        tries++;
        confd = connect(sockfd, serveraddr, serveraddr_size);
        if (confd < 0) { 
            printf("ERROR connecting to the Prover . Retrying attempt %d \n", tries);
        } else {
            printf("Successfully connect to Prover.\n");
            break;
        }
        
        if (tries == 4) {
            colorLog(KRED, "Unable to connect to the Prover after 4 attempts . Aborting...\n");
            colorLog(KRED, "Note: Make sure IP address is correct. (Value specified in config file\n)");
            exit(0);
        }
    }
 
    return confd;
}
