/* Copyright (C) 2017 Advanced Digital Sciences Centre
 * 
 * This is a software-based attestation demo code developed at Advanced Digital Sciences Center.
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * It is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * @author: Toby Zhou.Bin
 */

/* 
 * File:   Verifier.c
 * Author: Zhou Bin
 *
 * Created on 14 Sep, 2017, 4:20 PM
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <time.h>
#include "Utils.h"
#include "MemoryStride.h"

#define BUFSIZE 1024
#define IPSIZE 50
#define MESSAGE_SIZE 66
#define MESSAGE_HEADER_SIZE 6
#define MESSAGE_SEED_SIZE 52
#define MESSAGE_ITER_SIZE 8
#define SEED_NUMBER 13
#define SEED_SIZE 4
#define BASE_16 16
#define CHECKSUM_SIZE 48


void generateMsg(unsigned char *msg, unsigned long iterations) {
    int i=0; // for loop

    // This is fixed to create a IEC104 packet.
    unsigned char msgHeader[MESSAGE_HEADER_SIZE]= {0x68, 0x04, 0x03, 0x66, 0x69, 0x99};
    strncpy(msg, msgHeader, MESSAGE_HEADER_SIZE);

    // Generate Ramdon seed
    unsigned char elements[] = {'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A',
                                'B', 'C', 'D', 'E', 'F'};
    srand(time(NULL));
    for(i=0; i<MESSAGE_SEED_SIZE; i++)
        msg[MESSAGE_HEADER_SIZE + i] = elements[rand()%BASE_16];

    // Add iteration value
    unsigned char buffer[10];
    sprintf(buffer, "%08lX", iterations & 0xFFFFFFFF);
    //printf("8 bit: %s \n", buffer);
    for(i=0; i<MESSAGE_ITER_SIZE; i++)
        msg[MESSAGE_HEADER_SIZE + MESSAGE_SEED_SIZE + i] = buffer[i];

    // Ensure no unwanted tailing chars.
    msg[MESSAGE_SIZE] = '\0';
}


void getSeedIterFromMsg(unsigned char *msg, unsigned long *seeds, unsigned long *iteration) {
    int i=0, j=0; // for loop

    // msg = header + seeds + iteration
    unsigned char seed[SEED_SIZE + 1]; // +1 to store '\0'
    unsigned char iter[MESSAGE_ITER_SIZE+1];  // +1 to store '\0'

    for(i=0; i<SEED_NUMBER; i++) {
        for (j=0; j<SEED_SIZE; j++)
            seed[j] = msg[MESSAGE_HEADER_SIZE + i * SEED_SIZE + j];
        seed[SEED_SIZE] = '\0';
        seeds[i] = stringToHex(seed, SEED_SIZE);
    }

    // Get iteration
    for(i=0; i<MESSAGE_ITER_SIZE; i++)
        iter[i] = msg[MESSAGE_HEADER_SIZE + MESSAGE_SEED_SIZE + i];
    iter[MESSAGE_ITER_SIZE] = '\0';
    *iteration = stringToHex(iter, MESSAGE_ITER_SIZE);

}


int main(int argc, char *argv[]) {

    // Variables
    // Config
    char memoryImg[BUFSIZE], IP[IPSIZE];
    int portNum, timeThreshold;
    unsigned long iterations;
    // User input
    int option;
    // Socket & Connection
    int sockfd=-1, confd=-1;
    struct sockaddr_in serveraddr;
    struct hostent *prover;
    // Packets
    unsigned char message[MESSAGE_SIZE+1]; // +1 for '\0'
    int writeRes=-1, readRes=-1;
    unsigned char recvBuf[BUFSIZE];
    // Communication timing
    struct timeval start, end;
    float diff_time;
    int timeExceedLimit;
    // Memory Stride
    unsigned long seeds[SEED_NUMBER];
    unsigned long iteration;
    unsigned char simulatedMainRes[CHECKSUM_SIZE+1] = {0};
    // Results compare
    int sameResult;


    // Get config from file
    getVerifierConfig(memoryImg, IP, &portNum, &timeThreshold, &iterations);
    printf("Configs: %s\nIP:%s\nPort:%d\nTime(ms):%d\nIterations:%lu\n", memoryImg, IP, portNum, timeThreshold, iterations);


    // Get options from user
    colorLog(KYEL, "\n***** Welcome to Attestation Console *****\n");
    printf("Please select option:\n");
    printf("1. Send Memory Stride Packet\n");
    printf("2. Exit Program\n");
    scanf("%d", &option);
    if(option != 1) {
        printf("Exiting ... \n");
        exit(0);
    }


    // Socket & Connection
    printf("\n(0). Creating Socket & Connection\n");
    sockfd = createSocket();
    prover = findProver(IP);

    // build the prover's Internet address.
    bzero((char *) &serveraddr, sizeof (serveraddr));
    serveraddr.sin_family = AF_INET;
    bcopy((char *) prover->h_addr,
            (char *) &serveraddr.sin_addr.s_addr, prover->h_length);
    serveraddr.sin_port = htons(portNum);
    
    // Connect to prover.
    confd = connectProver(sockfd, (struct sockaddr *)&serveraddr, sizeof(serveraddr));

 
    // Generate packet
    printf("\n(1). Creating 'Attestation' Packet\n");
    generateMsg(message, iterations);
    printf("message: %s\nsize: %d\n", message, (int)strlen(message));


    // Send packet
    printf("\n(2). Sending Packet to Prover\n");
    gettimeofday(&start, NULL);
    writeRes = write(sockfd, message, MESSAGE_SIZE);
    if(writeRes < 0) {
        colorLog(KRED, "Fail to send packet. Aborting ... \n");
        exit(0);
    }


    // Receive the Prover's reply
    printf("\n(3). Receiving packet from Prover\n");
    readRes = read(sockfd, recvBuf, BUFSIZE);
    gettimeofday(&end, NULL);
    if(readRes < 0) {
        colorLog(KRED, "Fail to receive packet. Aborting ... \n");
        exit(0);
    }


    // Evaluate reply & timing
    printf("\n(4). Check result and timing\n");
    printf("Size of the response: ");
    printf("%s",KYEL);
    printf(" %d bytes \n", readRes);
    printf("%s",KNRM);

    // Size for checksum will be 48
    if(readRes != CHECKSUM_SIZE) {
        recvBuf[readRes] = '\0'; // remove unwanted tailing in recvBuf
        printf("This is not a checksum packet. The content is:\n");
        colorLog(KYEL, recvBuf);
        exit(0);
    }
    printf("Checksum computed by Prover: \n");
    printf("%s",KYEL);
    displayInHex(recvBuf, CHECKSUM_SIZE);
    printf("%s",KNRM);
 
    diff_time = timedifference_msec(start, end);
    printf("Execution time for checksum computation in milliseconds : ");
    printf("%s",KYEL);
    printf("%3.3f ms", diff_time);
    printf("%s\n",KNRM);
    if ((int)diff_time > timeThreshold ) {
        timeExceedLimit = 1;
        colorLog(KRED, "Execution time falls outside the expected range!!\n ");
    } else {
        timeExceedLimit = 0;
    }


    // Run Memory Stride
    printf("\n(5). Running the Memory Stride on Verifier\n");
    getSeedIterFromMsg(message, seeds, &iteration);
    // Check values
    int i=0; // for loop
    for(i=0; i<SEED_NUMBER; i++)
        printf("Seed %d: %02lx \n", i, seeds[i]);
    printf("Iteration: %lu\n", iteration);

    strideMemory(simulatedMainRes, memoryImg, seeds, iteration);
    simulatedMainRes[CHECKSUM_SIZE] = '\0';
    printf("Checksum computed by Verifier: \n");
    printf("%s",KYEL);
    displayInHex(simulatedMainRes, CHECKSUM_SIZE);
    printf("%s",KNRM);


    // Compare results.
    printf("\n(6). Compare results from Prover and Verifier\n");
    sameResult = 1;
    for(i=0; i<CHECKSUM_SIZE; i++) {
        if(recvBuf[i] != simulatedMainRes[i]){
            sameResult = 0;
            break;
        }
    }
    if(sameResult == 1)
        colorLog(KYEL, "The Verifier and Prover have SAME values!\n");
    else
        colorLog(KRED, "The Verifier and Prover have Different values!\n");



    // 7. Conclude
    printf("\n");
    if ((timeExceedLimit == 0)  && sameResult)
        colorLog(KYEL, "The Prover device is free from malware !\n");
    else
        colorLog(KRED, "The Prover device may be infected with malware !\n"); 
        


    return 0;
}